use core::future::*;
use core::pin::Pin;
use core::task::*;

pub enum DontAbortMode {
    PanicIfReused,
    HangIfReused,
}

pub struct DontAbort<A> {
    a: Option<A>,
    done: bool,
    mode: DontAbortMode,
}

pub struct DontAbortFuture<'a, A> {
    inner: &'a mut DontAbort<A>
}

impl<A> DontAbort<A> {
    pub fn new(a: A, mode: DontAbortMode) -> DontAbort<A> {
        DontAbort { a: Option::Some(a), done: false, mode }
    }

    pub fn hang() -> DontAbort<A> {
        DontAbort { a: Option::None, done: false, mode: DontAbortMode::HangIfReused }
    }

    #[allow(dead_code)]
    pub fn done(self: &Self) -> bool {
        return self.done;
    }

    pub fn continue_wait<'a>(self: &'a mut Self) -> DontAbortFuture<'a, A> {
        DontAbortFuture { inner: self }
    }
}

impl<A: Unpin> Unpin for DontAbort<A> {}
impl<'a, A: Unpin> Unpin for DontAbortFuture<'a, A> {}

impl<'a, A> Future for DontAbortFuture<'a, A>
where
    A: Future,
{
    type Output = A::Output;

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        let this = unsafe { self.get_unchecked_mut() };
        match (this.inner.done, this.inner.a.as_mut()) {
            (true, _) | (_, Option::None) =>
                match this.inner.mode {
                    DontAbortMode::PanicIfReused => panic!("Ready future is called again!"),
                    DontAbortMode::HangIfReused => return Poll::Pending,  // will never finish because we haven't scheduled ourselves again
                },
            (false, Option::Some(a)) => {
                let a = unsafe { Pin::new_unchecked(a) };
                if let Poll::Ready(x) = a.poll(cx) {
                    this.inner.done = true;
                    return Poll::Ready(x);
                }
                Poll::Pending
            }
        }
    }
}
